@@ -3,13 +3,35 @@ import logging
 import os
 
 from cardutil.config import config as pkg_config
-from cardutil import __version__
+from cardutil import __version__, CardutilError
+from cardutil.vendor.hexdump import hexdump
 
 LOGGER = logging.getLogger(__name__)
 
 
+def print_banner(command_name, parms):
+    message = f'{command_name} (cardutil {__version__})'
+    print(message)
+    print('parameters:')
+    for parm_key in parms:
+        if parms[parm_key]:
+            print(f' -{parm_key}:{parms[parm_key]}')
+
+
+def print_exception_details(err: CardutilError):
+    print("*** ERROR - processing has stopped ***")
+    if err.record_number:
+        print(f'Error detected in record {err.record_number}')
+    print(err)
+    if err.binary_context_data:
+        hexdump(err.binary_context_data)
+
+
 def add_version(parser):
-    parser.add_argument('--version', action='version', version=f'%(prog)s (cardutil {__version__})')
+    version_text = (f'%(prog)s (cardutil {__version__})\n'
+                    f'(C)Copyright 2019-2022 Anthony Delosa\n')
+
+    parser.add_argument('--version', action='version', version=version_text)
 
 
 def get_config(config_filename, envvar='CARDUTIL_CONFIG', cli_filename=None):
