@@ -67,16 +67,22 @@ import decimal
 import logging
 import re
 import struct
+import sys
 
+from cardutil import CardutilError
 from cardutil.BitArray import BitArray
 from cardutil.card import mask
 from cardutil.config import config
-from cardutil.hexdump import hexdump
+from cardutil.vendor.hexdump import hexdump
 
 LOGGER = logging.getLogger(__name__)
 DEFAULT_ENCODING = 'latin_1'
 
 
+class Iso8583DataError(CardutilError):
+    pass
+
+
 def dumps(obj: dict, encoding=None, iso_config=None, hex_bitmap=False):
     """
     Serialize obj to a ISO8583 message byte string
@@ -136,6 +142,9 @@ def _iso8583_to_dict(message, bit_config, encoding=DEFAULT_ENCODING, hex_bitmap=
 
     * Message Type indicator - 4 bytes
     * Binary bitmap - 16 bytes (Reads DE1 and DE2)
+    OR if hex_bitmap = True
+    * Message Type indicator - 4 bytes
+    * Hex bitmap - 32 bytes (Reads DE1 and DE2)
     * Message data - Remainder of record
 
     :param bit_config: dictionary of bit mapping configuration
@@ -152,21 +161,26 @@ def _iso8583_to_dict(message, bit_config, encoding=DEFAULT_ENCODING, hex_bitmap=
     # split raw message into components MessageType(4B), Bitmap(16B),
     # Message(l=*)
 
-    if hex_bitmap:
-        message_length = len(message)-36
-        message_type_indicator, bitmap, message_data = struct.unpack(
-            "4s32s" + str(message_length) + "s", message)
-        binary_bitmap = binascii.unhexlify(bitmap)
-
-    else:
-        message_length = len(message)-20
-        message_type_indicator, binary_bitmap, message_data = struct.unpack(
-            "4s16s" + str(message_length) + "s", message)
+    try:
+        if hex_bitmap:
+            message_length = len(message)-36
+            message_type_indicator, bitmap, message_data = struct.unpack(
+                "4s32s" + str(message_length) + "s", message)
+            binary_bitmap = binascii.unhexlify(bitmap)
 
+        else:
+            message_length = len(message)-20
+            message_type_indicator, binary_bitmap, message_data = struct.unpack(
+                "4s16s" + str(message_length) + "s", message)
+    except struct.error as ex:
+        raise Iso8583DataError('Failed unpacking bitmap values', binary_context_data=message, original_exception=ex)
     return_values = dict()
 
     # add the message type
-    return_values["MTI"] = message_type_indicator.decode(encoding)
+    try:
+        return_values["MTI"] = message_type_indicator.decode(encoding)
+    except UnicodeError as ex:
+        raise Iso8583DataError('Failed decoding MTI field', binary_context_data=message, original_exception=ex)
 
     message_pointer = 0
     bitmap_list = _get_bitmap_list(binary_bitmap)
@@ -186,12 +200,10 @@ def _iso8583_to_dict(message, bit_config, encoding=DEFAULT_ENCODING, hex_bitmap=
 
     # check that all of message has been consumed, otherwise raise exception
     if message_pointer != len(message_data):
-        raise ValueError(
-            "Message data not correct length. Bitmap indicates len={0}, message is len={1}\n{2}".format(
-                message_pointer,
-                len(message_data),
-                hexdump(message_data, result="return")
-            )
+        raise Iso8583DataError(
+            f'Message data not correct length. '
+            f'Bitmap indicates len={message_pointer}, message is len={len(message_data)}',
+            binary_context_data=message
         )
 
     return return_values
@@ -229,7 +241,7 @@ def _dict_to_iso8583(message, bit_config, encoding=DEFAULT_ENCODING, hex_bitmap=
         message[f'DE{de_field_key}'] = de_field_value
 
     for bit in range(2, 128):
-        if message.get('DE' + str(bit)):
+        if message.get('DE' + str(bit)) or message.get('DE' + str(bit)) == 0:  # 0 evals to false, allow zero values
             LOGGER.debug(f'processing bit {bit}')
             bitmap_values[bit - 1] = True
             LOGGER.debug(message.get('DE' + str(bit)))
@@ -283,16 +295,29 @@ def _iso8583_to_field(bit, bit_config, message_data, encoding=DEFAULT_ENCODING):
 
     if length_size > 0:
         field_length_string = message_data[:length_size]
-        field_length_string = field_length_string.decode(encoding)
-        field_length = int(field_length_string)
+        try:
+            field_length_string = field_length_string.decode(encoding)
+        except UnicodeDecodeError as ex:
+            raise Iso8583DataError(f'Unable to decode DE{bit} field length',
+                                   binary_context_data=message_data, original_exception=ex)
+
+        try:
+            field_length = int(field_length_string)
+        except ValueError as ex:
+            raise Iso8583DataError(f'Invalid field length DE{bit}',
+                                   binary_context_data=message_data, original_exception=ex)
 
     field_data = message_data[length_size:length_size + field_length]
     LOGGER.debug(f'field_data={field_data}')
-    field_processor = _get_parameter(bit_config, 'field_processor')
+    field_processor = bit_config.get('field_processor')
 
     # do ascii conversion except for ICC field
     if field_processor != 'ICC':
-        field_data = field_data.decode(encoding)
+        try:
+            field_data = field_data.decode(encoding)
+        except UnicodeDecodeError as ex:
+            raise Iso8583DataError(f'Unable to decode DE{bit} field value',
+                                   binary_context_data=message_data, original_exception=ex)
 
     # if field is PAN type, mask the card value
     if field_processor == 'PAN':
@@ -303,8 +328,11 @@ def _iso8583_to_field(bit, bit_config, message_data, encoding=DEFAULT_ENCODING):
         field_data = _pan_prefix(field_data)
 
     # do field conversion to native python type
-    field_data = _string_to_pytype(field_data, bit_config)
-
+    try:
+        field_data = _string_to_pytype(field_data, bit_config)
+    except ValueError as ex:
+        raise Iso8583DataError(f'Unable to convert DE{bit} field to python type',
+                               binary_context_data=message_data, original_exception=ex)
     return_values = dict()
 
     # add value to return dictionary
@@ -316,7 +344,8 @@ def _iso8583_to_field(bit, bit_config, message_data, encoding=DEFAULT_ENCODING):
 
     # if a DE43 field, break in down again and add to results
     if field_processor == 'DE43':
-        return_values.update(_get_de43_fields(field_data))
+        processor_config = bit_config.get('field_processor_config')
+        return_values.update(_get_de43_fields(field_data, processor_config))
 
     # if ICC field, break into tags
     if field_processor == 'ICC':
@@ -325,17 +354,6 @@ def _iso8583_to_field(bit, bit_config, message_data, encoding=DEFAULT_ENCODING):
     return return_values, field_length + length_size
 
 
-def _get_parameter(config, parameter):
-    """
-    Checks for parameter value and sets if present
-
-    :param config: bit configuration list
-    :param parameter: the configuration item to set
-    :return: string with value of parameter
-    """
-    return config[parameter] if config.get(parameter) else ""
-
-
 def _pan_prefix(field_data):
     """
     Get prefix of PAN
@@ -351,15 +369,15 @@ def _string_to_pytype(field_data, bit_config):
     :param bit_config: Configuration for bit
     :return: data in required type
     """
-    field_python_type = _get_parameter(bit_config, 'field_python_type')
+    field_python_type = bit_config.get('field_python_type')
+    field_date_format = bit_config.get('field_date_format', "%y%m%d")
 
     if field_python_type in ("int", "long"):
         field_data = int(field_data)
     if field_python_type == "decimal":
         field_data = decimal.Decimal(field_data)
     if field_python_type == "datetime":
-        field_data = datetime.datetime.strptime(
-            field_data, "%y%m%d%H%M%S")
+        field_data = datetime.datetime.strptime(field_data, field_date_format)
     return field_data
 
 
@@ -371,17 +389,55 @@ def _pytype_to_string(field_data, bit_config):
     :param bit_config: Configuration for bit
     :return: data in required type
     """
-    field_python_type = _get_parameter(bit_config, 'field_python_type')
+    field_python_type = bit_config.get('field_python_type')
+    field_date_format = bit_config.get('field_date_format', "%y%m%d")
+
     return_string = field_data
     if field_python_type in ('int', 'long'):
-        return_string = format(int(field_data), '0' + str(_get_parameter(bit_config, 'field_length')) + 'd')
+        return_string = format(int(field_data), '0' + str(bit_config.get('field_length', 0)) + 'd')
     if field_python_type == "decimal":
-        return_string = format(decimal.Decimal(field_data), '0' + str(_get_parameter(bit_config, 'field_length')) + 'f')
+        return_string = format(decimal.Decimal(field_data), '0' + str(bit_config.get('field_length', 0)) + 'f')
     if field_python_type == "datetime":
-        return_string = format(field_data, "%y%m%d%H%M%S")
+        if not isinstance(field_data, datetime.datetime):
+            field_data = _get_date_from_string(field_data)
+        return_string = format(field_data, field_date_format)
     return return_string
 
 
+def _get_date_from_string(field_data: str) -> datetime:
+    """
+    Parse string dates to python datetime object
+
+    Use dateutils library if it is installed, otherwise revert to simple parser
+    :param field_data: string containing date
+    :return: datetime object
+    """
+    try:
+        import dateutil.parser as parser
+        print('Using dateutil parser')
+        return parser.parse(field_data)
+    except ImportError:
+        pass
+
+    if sys.version_info >= (3, 7):
+        print('Using fromisoformat')
+        return datetime.datetime.fromisoformat(field_data)
+
+    # fallback parser -- tries a few different formats until one works
+    print('Using built in date parser')
+    date_formats = ["%Y-%m-%d %H:%M:%S", "%Y-%m-%d %H:%M", "%Y-%m-%d"]
+    output_date = None
+    for date_format in date_formats:
+        try:
+            output_date = datetime.datetime.strptime(field_data, date_format)
+            break
+        except ValueError:
+            continue
+    if not output_date:
+        raise ValueError("Unrecognised date string format - {}".format(field_data))
+    return output_date
+
+
 def _get_field_length(bit_config):
     """
     Determine length of iso8583 style field
@@ -426,11 +482,14 @@ def _pds_to_de(dict_values):
     :return: list of byte strings containing pds data, or None if no fields
     """
     # get the PDS field keys in order
+    LOGGER.debug(f'dict_values={dict_values}')
     keys = sorted([key for key in dict_values if key.startswith('PDS')])
+    LOGGER.debug(f'keys={keys}')
     output = ''
     outputs = []
     for key in keys:
         tag = int(key[3:])
+        LOGGER.debug(f'tag={tag}')
         length = len(dict_values[key])
         add_output = f'{tag:04}{length:03}{dict_values[key]}'
         if len(output + add_output) > 999:
@@ -522,7 +581,7 @@ def _icc_to_dict(field_data):
     return return_values
 
 
-def _get_de43_fields(de43_field):
+def _get_de43_fields(de43_field, processor_config=None):
     """
     get pds 43 field breakdown
 
@@ -530,18 +589,21 @@ def _get_de43_fields(de43_field):
     :return: dictionary of pds 43 sub elements
     """
     LOGGER.debug("de43_field=%s", de43_field)
-    de43_regex = (
-        r"(?P<DE43_NAME>.+?) *\\(?P<DE43_ADDRESS>.+?) *\\(?P<DE43_SUBURB>.+?) *\\"
-        r"(?P<DE43_POSTCODE>\S{4,10}) *(?P<DE43_STATE>.{3})(?P<DE43_COUNTRY>.{3})"
-    )
 
+    # No field config provided, just exit
+    if not processor_config:
+        return dict()
+
+    # perform regex field matching
+    de43_regex = processor_config
     field_match = re.match(de43_regex, de43_field)
     if not field_match:
         return dict()
 
     # get the dict
     field_dict = field_match.groupdict()
-
+    if field_dict.get('DE43_POSTCODE'):
+        field_dict['DE43_POSTCODE'] = field_dict['DE43_POSTCODE'].rstrip()
     return field_dict
 
 
