@@ -1,8 +1,11 @@
 """
 Mastercard |reg| IPM clearing file readers and writers
 
-* block and unblock 1014 blocked IPM and parameter files
-* process VBS format records
+* VBS file readers and writers
+* IPM file readers and writers
+* IPM parameter extract reader
+* Support for 1014 blocked format
+
 
 Read an IPM file::
 
@@ -33,7 +36,7 @@ A file consists of records. Each record is prefixed with a 4 byte binary length.
 
 Say you had a file with the following 2 records:
 
-.. code-block:: none
+.. code-block:: text
 
     "This is first record 1234567"  <- length 28
     "This is second record AAAABBBBB123"  <- length 34
@@ -88,12 +91,17 @@ In this case, there is only one increment
 import io
 import logging
 import struct
+import typing
 
-from cardutil import iso8583
+from cardutil import iso8583, config, CardutilError
 
 LOGGER = logging.getLogger(__name__)
 
 
+class MciIpmDataError(CardutilError):
+    pass
+
+
 class Block1014(object):
     """
     1014 Blocker for file objects.
@@ -105,7 +113,7 @@ class Block1014(object):
         self.file_obj = file_obj
         self.remaining_chars = 1012
 
-    def __getattr__(self, name):
+    def __getattr__(self, name: str):
         """
         Attribute proxy for wrapped file object
         """
@@ -116,7 +124,7 @@ class Block1014(object):
                 return getattr(self.file_obj, name)
         return None
 
-    def write(self, bytes_to_write):
+    def write(self, bytes_to_write: bytes) -> None:
         """
         Write requested bytes to the output file object.
         """
@@ -148,7 +156,7 @@ class Block1014(object):
         self.remaining_chars = 1012-len(bytes_to_write)
         LOGGER.debug(f'remaining_chars={self.remaining_chars}')
 
-    def seek(self, pos):
+    def seek(self, pos: int) -> None:
         """
         Finalise then seek file object to requested position
 
@@ -157,14 +165,14 @@ class Block1014(object):
         self.finalise()
         self.file_obj.seek(pos)
 
-    def close(self):
+    def close(self) -> None:
         """
         Finalise then close the file object
         """
         self.finalise()
         self.file_obj.close()
 
-    def finalise(self):
+    def finalise(self) -> None:
         """
         Complete the blocking operation by creating final 1014 block.
         Called by ``close`` and ``seek`` methods to ensure completion.
@@ -178,11 +186,11 @@ class Unblock1014(object):
     Unblocks 1014 blocked file objects.
     Wrap around a 1014 blocked file object. Return file like object providing only unblocked data
     """
-    def __init__(self, file_obj):
+    def __init__(self, file_obj: typing.BinaryIO):
         self.file_obj = file_obj
         self.buffer = b''
 
-    def __getattr__(self, name):
+    def __getattr__(self, name: str) -> any:
         """
         Attribute proxy for wrapped file object
         """
@@ -193,7 +201,7 @@ class Unblock1014(object):
                 return getattr(self.file_obj, name)
         return None
 
-    def read(self, bytes_to_read=0):
+    def read(self, bytes_to_read: int = 0):
         """
         Read requested bytes from the file object. Returned data will be unblocked
         """
@@ -222,7 +230,10 @@ class VbsReader(object):
                 print(vbs_record)
 
     """
-    def __init__(self, vbs_file, blocked=False):
+    record_number = 1
+    last_record = None
+
+    def __init__(self, vbs_file: typing.BinaryIO, blocked: bool = False):
         """
         Create a VbsReader object
 
@@ -232,7 +243,7 @@ class VbsReader(object):
         if blocked:
             self.vbs_data = Unblock1014(vbs_file)
 
-    def __getattr__(self, name):
+    def __getattr__(self, name) -> any:
         """
         Attribute proxy for wrapped file object
         """
@@ -246,20 +257,42 @@ class VbsReader(object):
     def __iter__(self):
         return self
 
-    def __next__(self):
+    def __next__(self) -> bytes:
         """
         Unpacks a variable blocked object into records
         """
         record_length_raw = self.vbs_data.read(4)
+        if len(record_length_raw) != 4:
+            # this can happen if the VBS does not have a zero length record at end.
+            # You can recreate using VbsWriter and not calling close method.
+            # The reader will just accept we are at end if this happens.
+            LOGGER.warning(f'Unable to read next record length - requested 4 bytes,'
+                           f' got {len(record_length_raw)} -- assuming end of data')
+            raise StopIteration
+
         record_length = struct.unpack(">i", record_length_raw)[0]
         LOGGER.debug("record_length=%s", record_length)
 
+        # throw mcipm data error if length is negative or excessively large (indicates bad input)
+        if record_length < 0 or record_length > 3000:
+            raise MciIpmDataError(f"Invalid record length - value read was {record_length}",
+                                  record_number=self.record_number,
+                                  binary_context_data=self.last_record)
+
         # exit if last record (length=0)
         if record_length == 0:
             raise StopIteration
 
         record = self.vbs_data.read(record_length)
-        return record
+        if len(record) != record_length:
+            raise MciIpmDataError(f"Unable to read complete record - record length: {record_length}, "
+                                  f"data read: {len(record)}",
+                                  record_number=self.record_number,
+                                  binary_context_data=record_length_raw + record)
+
+        self.last_record = record_length_raw + record  # save last record read
+        self.record_number += 1    # increment record counter
+        return record  # get the full record including the record length
 
 
 class IpmReader(VbsReader):
@@ -287,7 +320,8 @@ class IpmReader(VbsReader):
                 print(record)
 
     """
-    def __init__(self, ipm_file, encoding=None, iso_config=None, **kwargs):
+    def __init__(self, ipm_file: typing.BinaryIO,
+                 encoding: str = None, iso_config: dict = None, **kwargs):
         """
         Create a new IpmReader
 
@@ -299,10 +333,106 @@ class IpmReader(VbsReader):
         self.iso_config = iso_config
         super(IpmReader, self).__init__(ipm_file, **kwargs)
 
-    def __next__(self):
+    def __next__(self) -> dict:
+
         vbs_record = super(IpmReader, self).__next__()
         LOGGER.debug(f'{len(vbs_record)}: {vbs_record}')
-        return iso8583.loads(vbs_record, encoding=self.encoding, iso_config=self.iso_config)
+        try:
+            output = iso8583.loads(vbs_record, encoding=self.encoding, iso_config=self.iso_config)
+        except CardutilError as ex:
+            raise MciIpmDataError(
+                'Error while loading ISO8583 record',
+                binary_context_data=self.last_record,
+                record_number=self.record_number,
+                original_exception=ex
+            )
+        return output
+
+
+class IpmParamReader(VbsReader):
+    """
+    IPM Param reader can be used to iterate through an IPM parameter extract file.
+    The record is returned as a dictionary containing the parameter keys.
+
+    ::
+
+        from cardutil.mciipm import IpmParamReader
+        with open('param.bin', 'rb') as param_in:
+            reader = IpmParamReader(param_in, table_id='IP0040T1')
+            for record in reader:
+                print(record)
+
+    If the parameter file is 1014 block format, then set the ``blocked`` parameter to True.
+
+    ::
+
+        from cardutil.mciipm import IpmParamReader
+        with open('blocked_param.bin', 'rb') as param_in:
+            reader = IpmParamReader(param_in, table_id='IP0040T1', blocked=True)
+            for record in reader:
+                print(record)
+
+    """
+    # layout for IP0000T1
+    _IP0000T1_KEY = slice(11, 19)
+    _IP0000T1_TABLE_ID = slice(19, 27)
+    _IP0000T1_TABLE_SUB_ID = slice(243, 246)
+
+    # table type for all except IP0000T1 - get this 3 letter code from self.table_keys
+    _TABLE_SUB_ID = slice(8, 11)
+
+    def __init__(self, param_file: typing.BinaryIO, table_id: str, encoding: str = None, param_config: dict = None,
+                 **kwargs):
+        """
+        Create a new IpmParamReader
+
+        :param param_file: the file object to read
+        :param table_id: the IPM parameter table to read
+        :param encoding: the parameter file encoding
+        :param param_config: config dict with key bit_config
+        """
+        self.encoding = encoding if encoding else 'latin_1'
+        self.param_config = param_config if param_config else config.config.get('mci_parameter_tables')
+        self.table_id = table_id
+        self.table_index = dict()
+        super(IpmParamReader, self).__init__(param_file, **kwargs)
+
+        # check if config available for table id
+        if not self.param_config.get(table_id):
+            raise MciIpmDataError(f'Parameter config not available for table {table_id}')
+
+        # load the table index
+        trailer_record_found = False
+        while True:
+            try:
+                vbs_record = super(IpmParamReader, self).__next__()
+            except StopIteration:
+                break
+            record = vbs_record.decode(self.encoding)
+            if record[self._IP0000T1_KEY] == 'IP0000T1':
+                self.table_index[record[self._IP0000T1_TABLE_SUB_ID]] = record[self._IP0000T1_TABLE_ID]
+            if record.startswith('TRAILER RECORD IP0000T1'):
+                trailer_record_found = True
+                break
+        LOGGER.debug('IP0000T1 records: {}'.format(self.table_index))
+        if not trailer_record_found:
+            raise MciIpmDataError('parameter file missing IP0000T1 trailer record')
+
+    def __next__(self) -> dict:
+        while True:
+            record = super(IpmParamReader, self).__next__()
+            record_table_id = self.table_index.get(record[self._TABLE_SUB_ID].decode(self.encoding))
+            if record_table_id == self.table_id:
+                record_dict = dict()
+                for field in self.param_config[record_table_id]:
+                    record_dict[field] = self._get_param_field(record, field)
+                return record_dict
+
+    def _get_param_field(self, record, field):
+        table_id = self.table_index.get(record[self._TABLE_SUB_ID].decode(self.encoding))
+        return record[
+               self.param_config[
+                   table_id][field]["start"]:self.param_config[table_id][field]["end"]].decode(self.encoding)
 
 
 class VbsWriter(object):
@@ -316,17 +446,23 @@ class VbsWriter(object):
         ...     writer.write(b'This is the record')
         ...     writer.close()
 
-    The message is a byte string containing the data.
-
     The `close` method must be issued to finalise the file by adding the zero length record which
     indicated the end of the file.
+    The message is a byte string containing the data.
+
+    Alternatively, you can use as a context manager which will take care of the writer closure.
+
+        >>> with io.BytesIO() as vbs_out:
+        ...     with VbsWriter(vbs_out, blocked=True) as writer:
+        ...         writer.write(b'This is the record')
+
     """
-    def __init__(self, out_file, blocked=False):
+    def __init__(self, out_file: typing.BinaryIO, blocked: bool = False):
         self.out_file = out_file
         if blocked:
             self.out_file = Block1014(out_file)
 
-    def __getattr__(self, name):
+    def __getattr__(self, name) -> any:
         """
         Attribute proxy for wrapped file object
         """
@@ -337,7 +473,7 @@ class VbsWriter(object):
                 return getattr(self.out_file, name)
         return None
 
-    def write(self, record):
+    def write(self, record: bytes) -> None:
         """
         Add a new record to the VBS output file
 
@@ -353,7 +489,17 @@ class VbsWriter(object):
         # add data to output
         self.out_file.write(record)
 
-    def close(self):
+    def write_many(self, iterable: typing.Iterable[bytes]) -> None:
+        """
+        Convenience method to write multiple records from an iterable
+
+        :param iterable: iterable providing records as bytes
+        :return: None
+        """
+        for record in iterable:
+            self.write(record)
+
+    def close(self) -> None:
         """
         Finalise the VBS file output by adding the zero length file record.
 
@@ -363,6 +509,12 @@ class VbsWriter(object):
         self.out_file.write(struct.pack(">i", 0))
         self.out_file.seek(0)
 
+    def __enter__(self, *args, **kwargs):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
+        self.close()
+
 
 class IpmWriter(VbsWriter):
     """
@@ -376,7 +528,6 @@ class IpmWriter(VbsWriter):
         ...    writer.close()
 
     If the required file is 1014 block format, then set the ``blocked`` parameter to True.
-
     ::
 
         >>> with io.BytesIO() as ipm_out:
@@ -386,7 +537,6 @@ class IpmWriter(VbsWriter):
 
     You can provide the specific file encoding if required.
     All standard python encoding schemes are supported. Mainframe systems likely use ``cp500``
-
     ::
 
        >>> with io.BytesIO() as ipm_out:
@@ -394,8 +544,15 @@ class IpmWriter(VbsWriter):
        ...     writer.write({'MTI': '1111', 'DE2': '9999111122221111'})
        ...     writer.close()
 
+    Alternatively use as a context manager to ensure closure at end of processing
+    ::
+
+       >>> with io.BytesIO() as ipm_out:
+       ...     with IpmWriter(ipm_out, encoding='cp500') as writer:
+       ...         writer.write({'MTI': '1111', 'DE2': '9999111122221111'})
+
     """
-    def __init__(self, file_obj, encoding=None, iso_config=None, **kwargs):
+    def __init__(self, file_obj: typing.BinaryIO, encoding: str = None, iso_config: dict = None, **kwargs):
         """
         Create a new IpmWriter
 
@@ -406,7 +563,7 @@ class IpmWriter(VbsWriter):
         self.iso_config = iso_config
         super(IpmWriter, self).__init__(file_obj, **kwargs)
 
-    def write(self, obj):
+    def write(self, obj: dict) -> None:
         """
         Writes new record to IPM file
 
@@ -419,8 +576,18 @@ class IpmWriter(VbsWriter):
         record = iso8583.dumps(obj, encoding=self.encoding, iso_config=self.iso_config)
         super(IpmWriter, self).write(record)
 
+    def write_many(self, iterable: typing.Iterable[dict]) -> None:
+        """
+        Convenience method to write multiple records from an iterable
 
-def unblock_1014(input_data, output_data):
+        :param iterable: iterable providing records as dict
+        :return: None
+        """
+        for record in iterable:
+            self.write(record)
+
+
+def unblock_1014(input_data: typing.BinaryIO, output_data: typing.BinaryIO):
     """
     Unblocks a 1014 byte blocked file object
 
@@ -433,15 +600,15 @@ def unblock_1014(input_data, output_data):
         if not record:
             break
         if len(record) != 1014:
-            raise ValueError('Invalid file size')
+            raise MciIpmDataError('Invalid record size for 1014 blocked')
         if record[-2:] != pad_char * 2:
-            raise ValueError('Invalid 1014 block line ending')
+            raise MciIpmDataError('Invalid line ending for 1014 blocked')
         output_data.write(record[0:1012])
     output_data.seek(0)
     input_data.seek(0)
 
 
-def block_1014(input_data, output_data):
+def block_1014(input_data: typing.BinaryIO, output_data: typing.BinaryIO):
     """
     Creates a 1014 byte blocked file object
 
@@ -463,24 +630,32 @@ def block_1014(input_data, output_data):
     input_data.seek(0)
 
 
-def vbs_list_to_bytes(byte_list: list, blocked: bool = False) -> bytes:
+def vbs_list_to_bytes(byte_list: iter, **kwargs) -> bytes:
     """
     Convenience function for creating VBS byte strings (optionally blocked) from list of byte strings
+
+    :param byte_list: a list containing byte string records
+    :param kwargs: any options to be passed to VbsWriter constructor. See :py:mod:`cardutil.mciipm.VbsWriter`
+    :return: single byte string containing VBS data.
     """
     file_out = io.BytesIO()
-    vbs_out = VbsWriter(file_out, blocked=blocked)
+    vbs_out = VbsWriter(file_out, **kwargs)
     for rec in byte_list:
         vbs_out.write(rec)
     vbs_out.close()
     return file_out.read()
 
 
-def vbs_bytes_to_list(vbs_bytes: bytes, blocked: bool = False) -> list:
+def vbs_bytes_to_list(vbs_bytes: bytes, **kwargs) -> list:
     """
     Convenience function for unpacking VBS byte strings to byte string list
+
+    :param vbs_bytes: single byte string containing VBS data
+    :param kwargs: any options to be passed to VbsReader constructor. See :py:mod:`cardutil.mciipm.VbsReader`
+    :return: a list containing byte string records
     """
     file_in = io.BytesIO(vbs_bytes)
-    return [record for record in VbsReader(file_in, blocked=blocked)]
+    return [record for record in VbsReader(file_in, **kwargs)]
 
 
 if __name__ == '__main__':
