@@ -18,8 +18,8 @@ def test_message(encoding='ascii', hex_bitmap=False):
         '1144'.encode(encoding) +
         bitmap +
         ('164444555544445555111111000000009999150815171500123456789012333123423579957991200000'
-         '012306120612345612345657994211111111145BIG BOBS\\80 KERNDALE ST\\DANERLEY\\3103  VIC'
-         'AUS0080001001Y99901600000000000000011234567806999999').encode(encoding))
+         '012306120612345612345657994211111111149BIG BOBS\\80 KERNDALE ST\\DANERLEY\\3103      '
+         'VICAUS0080001001Y99901600000000000000011234567806999999').encode(encoding))
 
 
 message_ascii_raw = test_message()
