@@ -3,10 +3,11 @@ import datetime
 import decimal
 import unittest
 
+from cardutil import CardutilError
 from cardutil.config import config
 from cardutil.iso8583 import (
     BitArray, _iso8583_to_field, _field_to_iso8583, _iso8583_to_dict, _dict_to_iso8583, loads, dumps,
-    _pds_to_de, _pds_to_dict, _pytype_to_string, _icc_to_dict, _get_de43_fields)
+    _pds_to_de, _pds_to_dict, _pytype_to_string, _icc_to_dict, _get_de43_fields, _get_date_from_string)
 
 from tests import message_ebcdic_raw, message_ascii_raw, message_ascii_raw_hex, message_ebcdic_raw_hex
 
@@ -21,6 +22,13 @@ class Iso8583TestCase(unittest.TestCase):
         out_data = dumps({'MTI': '1234', 'DE2': '123'}, iso_config=config["bit_config"])
         self.assertEqual(b'1234\xc0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0003123', out_data)
 
+    def test_dumps_with_no_values(self):
+        # check that empty fields are not added as elements to ISO output
+        in_data_empty = {'MTI': '1234', 'DE2': '', 'DE3': None, 'DE4': 0}
+        out_data_empty = loads(dumps(in_data_empty))
+        print(out_data_empty)
+        self.assertEqual({'MTI': '1234', 'DE4': 0}, out_data_empty)
+
     def test_dumps_hex_bitmap(self):
         out_data = dumps({'MTI': '1234', 'DE2': '123'}, hex_bitmap=True)
         self.assertEqual(b'1234c000000000000000000000000000000003123', out_data)
@@ -92,6 +100,22 @@ class Iso8583TestCase(unittest.TestCase):
             ({'DE1': decimal.Decimal('123.432')}, 20),
             _iso8583_to_field('1', {'field_type': 'FIXED', 'field_python_type': 'decimal', 'field_length': 20},
                               b'0000000000000123.432'))
+        # trigger decode error field length
+        with self.assertRaises(CardutilError):
+            _iso8583_to_field('1', {'field_type': 'LLVAR', 'field_length': 0}, b'\x00\x004564320012321122')
+        # unable to decode field length
+        with self.assertRaises(CardutilError):
+            _iso8583_to_field('01', {"field_type": "LLLVAR", "field_length": 255}, b'01\xff', encoding='ascii')
+        # invalid field length
+        with self.assertRaises(CardutilError):
+            _iso8583_to_field('01', {"field_type": "LLLVAR", "field_length": 255}, b'01X.', encoding='ascii')
+        # unable to decode field value
+        with self.assertRaises(CardutilError):
+            _iso8583_to_field('01', {"field_type": "LLLVAR", "field_length": 255}, b'001\xff', encoding='ascii')
+        # unable to convert to python value
+
+        with self.assertRaises(CardutilError):
+            _iso8583_to_field('1', {'field_type': 'LLVAR', 'field_python_type': 'int', 'field_length': 0}, b'04XXXX')
 
     def test_field_to_iso8583(self):
         self.assertEqual(b'164564320012321122', _field_to_iso8583({'field_type': 'LLVAR'}, "4564320012321122"))
@@ -124,7 +148,8 @@ class Iso8583TestCase(unittest.TestCase):
                 decimal.Decimal("123.432")))
         self.assertEqual(
             b'140102151610', _field_to_iso8583(
-                {'field_type': 'FIXED', 'field_python_type': 'datetime', 'field_length': 12},
+                {'field_type': 'FIXED', 'field_python_type': 'datetime',
+                 'field_length': 12, "field_date_format": "%y%m%d%H%M%S"},
                 datetime.datetime(2014, 1, 2, 15, 16, 10)))
 
     def test_iso8583_to_dict(self):
@@ -132,7 +157,7 @@ class Iso8583TestCase(unittest.TestCase):
                          'DE12': datetime.datetime(2015, 8, 15, 17, 15, 0),
                          'DE22': '123456789012', 'DE24': '333', 'DE26': 1234,
                          'DE31': '57995799120000001230612', 'DE33': '123456', 'DE38': '123456',
-                         'DE42': '579942111111111', 'DE43': 'BIG BOBS\\80 KERNDALE ST\\DANERLEY\\3103  VICAUS',
+                         'DE42': '579942111111111', 'DE43': 'BIG BOBS\\80 KERNDALE ST\\DANERLEY\\3103      VICAUS',
                          'DE43_NAME': 'BIG BOBS', 'DE43_ADDRESS': '80 KERNDALE ST', 'DE43_SUBURB': 'DANERLEY',
                          'DE43_POSTCODE': '3103', 'DE43_STATE': 'VIC', 'DE43_COUNTRY': 'AUS',
                          'DE48': '0001001Y', 'PDS0001': 'Y', 'DE49': '999', 'DE63': '0000000000000001',
@@ -144,9 +169,18 @@ class Iso8583TestCase(unittest.TestCase):
         self.assertEqual(ebcdic_dict, expected_dict)
 
         # check exception when full message not processed
-        with self.assertRaises(ValueError):
+        with self.assertRaises(CardutilError):
             _iso8583_to_dict(message_ascii_raw + b' ', config["bit_config"], "ascii")
 
+        # check exception when cannot unpack mti and bitmap (struct error)
+        with self.assertRaises(CardutilError):
+            _iso8583_to_dict(b'\x00TI0', config["bit_config"], "ascii")
+
+        # check exception when cannot decode MTI
+        with self.assertRaises(CardutilError):
+            _iso8583_to_dict(b'\xFFBCDFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF.....', config["bit_config"],
+                             encoding="ascii", hex_bitmap=True)
+
     def test_dict_to_iso8583(self):
         source_dict = {'MTI': '1144', 'DE2': '4444555544445555', 'DE3': '111111', 'PDS0001': '1', 'PDS9999': 'Z'}
         actual_iso = _dict_to_iso8583(source_dict, config['bit_config'])
@@ -180,8 +214,8 @@ class Iso8583TestCase(unittest.TestCase):
         self.assertEqual('00001', _pytype_to_string('1', {'field_python_type': 'int', 'field_length': 5}))
         self.assertEqual('00001', _pytype_to_string('1', {'field_python_type': 'long', 'field_length': 5}))
         self.assertEqual(
-            '180101171500',
-            _pytype_to_string(datetime.datetime(2018, 1, 1, 17, 15), {'field_python_type': 'datetime'}))
+            '180101171500', _pytype_to_string(datetime.datetime(2018, 1, 1, 17, 15), {
+                'field_python_type': 'datetime', 'field_date_format': '%y%m%d%H%M%S'}))
 
     def test_icc_to_dict(self):
         self.assertEqual(
@@ -201,7 +235,91 @@ class Iso8583TestCase(unittest.TestCase):
         print(_icc_to_dict(test_de55))
 
     def test_get_de43_fields(self):
-        self.assertEqual(_get_de43_fields('THIS DOES NOT MATCH REGEX'), {})
+        default_processor_config = config['bit_config']['43'].get('field_processor_config')
+        # does not match
+        self.assertEqual(_get_de43_fields('THIS DOES NOT MATCH REGEX', default_processor_config), {})
+        # matches regex
+        expected_dict = {'DE43_ADDRESS': '100 MAIN ST',
+                         'DE43_COUNTRY': 'AUS',
+                         'DE43_NAME': 'BOBS BURGERS',
+                         'DE43_POSTCODE': '4102',
+                         'DE43_STATE': 'QLD',
+                         'DE43_SUBURB': 'WOOLLOONGABBA'}
+        self.assertEqual(
+            _get_de43_fields(
+                'BOBS BURGERS\\100 MAIN ST\\WOOLLOONGABBA\\4102      QLDAUS', default_processor_config), expected_dict)
+        custom_processor_config = r'(?P<DE43_ALL>.*)'
+        self.assertEqual(
+            _get_de43_fields('ALL FIELD', processor_config=custom_processor_config), {'DE43_ALL': 'ALL FIELD'})
+
+    def test_get_de43_fields_international_addresses(self):
+        default_processor_config = config['bit_config']['43'].get('field_processor_config')
+        # United Kingdom
+        de43_gbr = _get_de43_fields('The king\\Buckingham Palace\\LONDON\\SW1A 1AA     GBR', default_processor_config)
+        self.assertEqual(de43_gbr['DE43_POSTCODE'], 'SW1A 1AA')
+        self.assertEqual(de43_gbr['DE43_STATE'], '   ')
+        self.assertEqual(de43_gbr['DE43_COUNTRY'], 'GBR')
+
+        # Ireland
+        de43_irl = _get_de43_fields('Guinness\\St James Gate\\DUBLIN\\D08 VF8H     IRL', default_processor_config)
+        self.assertEqual(de43_irl['DE43_POSTCODE'], 'D08 VF8H')
+        self.assertEqual(de43_irl['DE43_STATE'], '   ')
+        self.assertEqual(de43_irl['DE43_COUNTRY'], 'IRL')
+
+    def test_get_date_from_string_use_fromisodate(self):
+        import builtins
+        import sys
+
+        # skip test if being run on py36 or less
+        if sys.version_info < (3, 7):
+            self.skipTest('Needs py37 or greater to run')
+
+        real_import = builtins.__import__
+
+        def mock_import(name, *vars):
+            if name == 'dateutil.parser':
+                raise ImportError
+            else:
+                return real_import(name, *vars)
+
+        builtins.__import__ = mock_import
+        self.get_date_from_string()
+        builtins.__import__ = real_import
+
+    def test_get_date_from_string_use_builtin(self):
+        import sys
+        import builtins
+
+        real_version = sys.version_info
+        sys.version_info = (3, 6)
+        real_import = builtins.__import__
+
+        def mock_import(name, *vars):
+            if name == 'dateutil.parser':
+                raise ImportError
+            else:
+                return real_import(name, *vars)
+
+        builtins.__import__ = mock_import
+        self.get_date_from_string()
+        with self.assertRaises(ValueError):
+            _get_date_from_string("11221-11-22")
+        builtins.__import__ = real_import
+        sys.version_info = real_version
+
+    def test_get_date_from_string_use_dateutil(self):
+        try:
+            import dateutil
+        except ImportError:
+            self.skipTest('requires that python-dateutil is installed')
+            return
+        print(dateutil)
+        self.get_date_from_string()
+
+    def get_date_from_string(self):
+        self.assertEqual(_get_date_from_string("2002-01-01"), datetime.datetime(2002, 1, 1))
+        self.assertEqual(_get_date_from_string("2002-01-01 10:01"), datetime.datetime(2002, 1, 1, 10, 1))
+        self.assertEqual(_get_date_from_string("2002-01-01 10:01:02"), datetime.datetime(2002, 1, 1, 10, 1, 2))
 
 
 if __name__ == '__main__':
